
import urllib
import xml.dom.minidom
import time, string
import os, math
import pickle

dbNodes = []
x = 0

class QuadTreeBin:
	def __init__(self, parentIn = None):
		if parentIn == None:
			self.tl = (90.0,-180.0)
			self.br = (-90.0,180.0)
		self.parent = parentIn

	def InitDecimalCalc(self):
		dlat = self.GetTL()[0] - self.GetBR()[0]
		dlon = self.GetBR()[1] - self.GetTL()[1]
		minres = 0.0000001
		maxbits = 16
		maxrange = pow(2,maxbits)

		latbitsRequired = int(math.ceil(math.log(dlat/minres,2)))
		lonbitsRequired = int(math.ceil(math.log(dlon/minres,2)))
		bitsRequired = latbitsRequired
		if lonbitsRequired > bitsRequired:
			bitsRequired = lonbitsRequired

		if bitsRequired < maxbits:
			self.decimalBits = bitsRequired
		#print latbitsRequired, lonbitsRequired
		#exit(0)
		#if dlat/minres < maxrange and dlon/minres < maxrange:
		#	self.decimalBits = maxbits

	def BinNodes(self, depth = 0):
		if not hasattr(self,"bins"):
			raise Exception('no child bins to use')
		for (nid,version) in self.GetOwnNodeIds():
			#print nid,version
			npos = self.GetOwnNodeByIdVer(nid,version)
			#Find which bin to use
			found = 0
			for bin in self.bins:
				btl = bin.GetTL()
				bbr = bin.GetBR()
				if npos[1] >= btl[1] and npos[1] <= bbr[1] \
					and npos[0] <= btl[0] and npos[0] >= bbr[0]:
					bin.AddNode(nid, version, npos,depth+1)
					found = 1
			if found == 0:
				raise Exception('bin not found for node')

		self.ClearOwnNodes()

	def CountBins(self):
		if hasattr(self,"bins"):
			count = 0
			for bin in self.bins:
				count = count + bin.CountBins()
			return count + 1
		return 1

	def GetTL(self):
		if self.parent == None:
			if not hasattr(self,"tl"):
				raise Exception('top parent should have bounds set')
			return self.tl
		binId = self.parent.bins.index(self)

		if not hasattr(self.parent,"cent"):
			self.parent.cent = (0.5 * self.parent.GetTL()[0] + 0.5 * self.parent.GetBR()[0],\
				0.5 * self.parent.GetTL()[1] + 0.5 * self.parent.GetBR()[1])
		if binId == 0: return self.parent.GetTL()
		if binId == 1: return (self.parent.GetTL()[0],self.parent.cent[1])
		if binId == 2: return (self.parent.cent[0],self.parent.GetTL()[1])
		if binId == 3: return self.parent.cent

		raise Exception('unexpected index of bin')

	def GetBR(self):
		if self.parent == None:
			if not hasattr(self,"br"):
				raise Exception('top parent should have bounds set')
			return self.br

		binId = self.parent.bins.index(self)

		if not hasattr(self.parent,"cent"):
			self.parent.cent = (0.5 * self.parent.GetTL()[0] + 0.5 * self.parent.GetBR()[0],\
				0.5 * self.parent.GetTL()[1] + 0.5 * self.parent.GetBR()[1])

		if binId == 0: return self.parent.cent
		if binId == 1: return (self.parent.cent[0],self.parent.GetBR()[1])
		if binId == 2: return (self.parent.GetBR()[0],self.parent.cent[1])
		if binId == 3: return self.parent.GetBR()

		raise Exception('unexpected index of bin')

	def GetOwnNodeIds(self):
		if hasattr(self, "nodes"):
			ids = self.nodes.keys()
			if hasattr(self, "dnodes"):
				ids.extend(self.dnodes.keys())
			#print ids
			return ids
		if hasattr(self, "dnodes"):
			ids = self.dnodes.keys()
			#print ids
			return ids
		return []

	def GetOwnNodeByIdVer(self, nid, version):
		if hasattr(self, "nodes"):
			if (nid,version) in self.nodes:
				#print "a",self.nodes[(nid,version)],(nid,version)
				return self.nodes[(nid,version)]
		if hasattr(self, "dnodes"):
			if (nid,version) in self.dnodes:
				maxrange = pow(2,self.decimalBits)
				pos = self.dnodes[(nid,version)]
				nlat = (float(pos[0]) / maxrange) * (self.GetTL()[0] - self.GetBR()[0]) + self.GetBR()[0]
				nlon = (float(pos[1]) / maxrange) * (self.GetBR()[1] - self.GetTL()[1]) + self.GetTL()[1]
				return (nlat, nlon)
		return None
	def ClearOwnNodes(self):
		if hasattr(self, "nodes"):
			self.nodes.clear()
		if hasattr(self, "dnodes"):
			self.dnodes.clear()
	def SetOwnNode(self, nid, version, pos):
		#print nid,version
		if hasattr(self, "decimalBits"):
			maxrange = pow(2,self.decimalBits)
			declat = int(maxrange * (pos[0] - self.GetBR()[0]) / (self.GetTL()[0] - self.GetBR()[0]))
			declon = int(maxrange * (pos[1] - self.GetTL()[1]) / (self.GetBR()[1] - self.GetTL()[1]))
			#print (declat, declon)
			if not hasattr(self, "dnodes"):
				self.dnodes = {}
			self.dnodes[(nid,version)] = (declat, declon)
		else:
			if not hasattr(self, "nodes"):
				self.nodes = {}
			self.nodes[(nid,version)] = pos

	def InformParentNodeInBin(self, nid, version, bin):
		if self.parent is not None:
			self.parent.InformParentNodeInBin(nid, version, bin)
		if hasattr(self, "nodeInBinCallbacks"):
			#print "InformParentNodeInBin", nid, version, bin
			for callback in self.nodeInBinCallbacks:
				callback(nid, version, bin)

	def AddNode(self, nid, version, pos, depth = 0):
		#print nid,version
		self.SetOwnNode(nid,version,pos)
		self.InformParentNodeInBin(nid,version,self)
		maxDepth = 30

		if len(self.GetOwnNodeIds()) > 100 and depth < maxDepth:
			#Split quad bin
			self.bins = (QuadTreeBin(self), QuadTreeBin(self), QuadTreeBin(self), QuadTreeBin(self))
			for bin in self.bins:
				bin.InitDecimalCalc()		

		if hasattr(self,"bins"):
			self.BinNodes(depth)

	def GetNodeByIdExhaustive(self,nid,version):
		#print "GetNodeByIdExhaustive",self.GetTL(), self.GetBR()
		if (nid,version) in self.GetOwnNodeIds():
			return self.GetOwnNodeByIdVer(nid,version)

		if hasattr(self,"bins"):
			for bin in self.bins:
				find = bin.GetNodeByIdExhaustive(nid,version)
				if find is not None:
					return find

	def GetBinContainingNodeByIdExhaustive(self,nid,version):
		#print "GetNodeByIdExhaustive",self.GetTL(), self.GetBR()
		if (nid,version) in self.GetOwnNodeIds():
			return self

		if hasattr(self,"bins"):
			for bin in self.bins:
				find = bin.GetBinContainingNodeByIdExhaustive(nid,version)
				if find is not None:
					return find

		return None

class NodePosStoreQuadTree():
	def __init__(self):
		self.quadtree = QuadTreeBin()
		self.quadtree.nodeInBinCallbacks = [self.NodeInBinFunc]
		self.idmap = {}

	def NodeInBinFunc(self, nid, version, bin):
		self.idmap[(nid, version)] = bin

	def InsertNode(self, nid, version, pos):
		return self.quadtree.AddNode(nid, version, pos)
	def GetNode(self, nid, version):
		if (nid, version) in self.idmap:
			bin = self.idmap[(nid, version)]
			return bin.GetOwnNodeByIdVer(nid, version)

		#return self.quadtree.GetNodeByIdExhaustive(nid,version)
	def GetNodes(self, bbox):
		(left,bottom,right,top) = bbox

	def Save(self, finame):
		fi = open(finame,"w")
		self.quadtree.nodeInBinCallbacks = []
		pickle.dump(self.quadtree, fi)
		self.quadtree.nodeInBinCallbacks = [self.NodeInBinFunc]
	def Load(self, finame):
		fi = open(finame,"r")
		self.quadtree = pickle.load(fi)
		self.quadtree.nodeInBinCallbacks = [self.NodeInBinFunc]

class NodePosStoreDict():
	def __init__(self):
		self.idmap = {}

	def InsertNode(self, nid, version, pos):
		self.idmap[(nid, version)] = pos
	def GetNode(self, nid, version):
		if (nid, version) in self.idmap:
			return self.idmap[(nid, version)]
	def GetNodes(self, bbox):
		(left,bottom,right,top) = bbox

	def Save(self, finame):
		fi = open(finame,"w")
		pickle.dump(self.idmap, fi)
		pass
	def Load(self, finame):
		fi = open(finame,"r")
		self.idmap = pickle.load(fi)
		pass

nodePos = NodePosStoreQuadTree()
#nodePos = NodePosStoreDict()

def ProcessChange(elements):
	for ch in elements:
		chNode = ch.childNodes
		for el in chNode:
			if el.nodeType is not el.ELEMENT_NODE:
				continue
			if el.hasAttribute("id"):
				idNum = int(el.attributes["id"].value)
				verNum = int(el.attributes["version"].value)
				#print el.tagName, idNum, verNum

				if el.tagName != "node": continue

				lat = float(el.attributes["lat"].value)
				lon = float(el.attributes["lon"].value)
				nodePos.InsertNode(idNum, verNum, (lat, lon))
				#print lat, lon

				#out.append((el.tagName, int(idNum)))
				if int(idNum) not in dbNodes:
					dbNodes.append(int(idNum))
				#if int(idNum) not in dbNodes:
				#	dbNodes[int(idNum)] = {}
				#dbNodes[int(idNum)][int(verNum)] = el

				

for root, dirs, files in os.walk('changesets'):
	count = 0
	print root, len(files)
	numFiles = len(files)
	for finame in files:
		finamejoin = os.path.join(root, finame)
		print finamejoin, count, numFiles
		fi = open(finamejoin,"r")

		try:
			xmldom = xml.dom.minidom.parseString(fi.read())
		except xml.parsers.expat.ExpatError:
			print "Deleting invalid OSM/XML file",finamejoin
			fi.close()
			os.remove(finamejoin)
			continue
	
		rootdoc = xmldom.getElementsByTagName("osmChange")[0]
		modify = rootdoc.getElementsByTagName("modify")
		delete = rootdoc.getElementsByTagName("delete")
		create = rootdoc.getElementsByTagName("create")

		ProcessChange(create)
		ProcessChange(modify)
		ProcessChange(delete)
		#print len(dbNodes)

		if hasattr(nodePos,"quadtree"):
			print nodePos.quadtree.CountBins()
		count = count + 1
		#if count > 30: break

	#print quadtree.GetNodeByIdExhaustive(668048580)
	print "query 762563303",nodePos.GetNode(762563303,1)
	#print "test",(nodePos.quadtree.GetBinContainingNodeByIdExhaustive(762563303,1).GetOwnNodeIds())
	print "idmap",len(nodePos.idmap)

nodePos.Save("nodepos.dat")
print "All done!"


